/*
 * Copyright 2010-2012 Craig S. Dickson (http://craigsdickson.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.squeakysand.jsp.tldgenerator;

import com.squeakysand.commons.lang.ToStringHelper;
import com.squeakysand.schema.javaee.ObjectFactory;
import com.squeakysand.schema.javaee.PathType;
import com.squeakysand.schema.javaee.TagFileType;
import com.squeakysand.schema.javaee.TldCanonicalNameType;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.codehaus.plexus.util.FileUtils;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class TagFileTypeListBuilder {

    private static final Logger LOG = LoggerFactory.getLogger(TagFileTypeListBuilder.class);

    private ObjectFactory objectFactory;
    private File tagFilesDir;
    private String tagFilesInclude;
    private String tagFilesExclude;

    public TagFileTypeListBuilder() {
        objectFactory = new ObjectFactory();
    }

    public File getTagFilesDir() {
        return tagFilesDir;
    }

    public void setTagFilesDir(File tagFilesDir) {
        this.tagFilesDir = tagFilesDir;
    }

    public String getTagFilesExclude() {
        return tagFilesExclude;
    }

    public void setTagFilesExclude(String tagFilesExclude) {
        this.tagFilesExclude = tagFilesExclude;
    }

    public String getTagFilesInclude() {
        return tagFilesInclude;
    }

    public void setTagFilesInclude(String tagFilesInclude) {
        this.tagFilesInclude = tagFilesInclude;
    }

    public List<TagFileType> build() throws TldGeneratorException {
        LOG.debug("tagFilesDir: {}, tagFilesInclude: {}, tagFilesExclude: {}", new Object[] {tagFilesDir, tagFilesInclude, tagFilesExclude});
        List<TagFileType> result = new ArrayList<TagFileType>();
        List<File> files = new ArrayList<File>();
        if (tagFilesDir.exists()) {
            try {
                files = FileUtils.getFiles(tagFilesDir, tagFilesInclude, tagFilesExclude);
                LOG.debug("files: {}", ToStringHelper.toString(files));
            } catch (IOException e) {
                throw new TldGeneratorException(e);
            }
        }
        for (File file : files) {
            TagFileType tagFile = objectFactory.createTagFileType();
            // set the name
            String name = generateDefaultTagName(file);
            TldCanonicalNameType xmlName = objectFactory.createTldCanonicalNameType();
            xmlName.setValue(name);
            tagFile.setName(xmlName);
            // set the path
            String path = generateTagFilePath(file);
            PathType xmlPath = objectFactory.createPathType();
            xmlPath.setValue(path);
            tagFile.setPath(xmlPath);
            // add the tag to the list of existing tags
            result.add(tagFile);
        }
        return result;
    }

    private String generateDefaultTagName(File file) {
        String name = file.getName();
        // lowercase the first letter
        name = Character.toLowerCase(name.charAt(0)) + name.substring(1);
        // trim off the .tag file extension
        if (name.endsWith(".tag")) {
            name = name.substring(0, name.length() - 4);
        } else if (name.endsWith(".tagx")) {
            // tag file name ends with .tagx instead
            name = name.substring(0, name.length() - 5);
        } else {
            throw new IllegalArgumentException(String.format("the extension of file %s must be either .tag or .tagx", file));
        }
        return name;
    }

    private String generateTagFilePath(File file) {
        String fullFilePath = file.getPath();
        String tagFilesDirPath = tagFilesDir.getPath();
        LOG.debug("tagFilesDirPath is: " + tagFilesDirPath);
        int index = tagFilesDirPath.length();
        LOG.debug("index is " + index);
        LOG.debug("File.separator is: " + File.separator);
        if (!tagFilesDirPath.endsWith(File.separator)) {
            index++;
            LOG.debug("index is now: " + index);
        }
        return "/META-INF/tags/" + fullFilePath.substring(index);
    }

}
