/*
 * Copyright 2010-2012 Craig S. Dickson (http://craigsdickson.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.squeakysand.jsp.tldgenerator;

import com.squeakysand.commons.lang.CompareToHelper;
import com.squeakysand.commons.lang.StringUtils;
import com.squeakysand.javadoc.JavaDocHelper;
import com.squeakysand.jsp.tagext.TagLibraryUtils;
import com.squeakysand.jsp.tagext.annotations.JspTag;
import com.squeakysand.schema.javaee.BodyContentType;
import com.squeakysand.schema.javaee.DescriptionType;
import com.squeakysand.schema.javaee.FullyQualifiedClassType;
import com.squeakysand.schema.javaee.GenericBooleanType;
import com.squeakysand.schema.javaee.JavaIdentifierType;
import com.squeakysand.schema.javaee.ObjectFactory;
import com.squeakysand.schema.javaee.TagType;
import com.squeakysand.schema.javaee.TldAttributeType;
import com.squeakysand.schema.javaee.TldCanonicalNameType;
import com.squeakysand.schema.javaee.VariableType;
import com.squeakysand.schema.javaee.XsdStringType;
import com.sun.javadoc.AnnotationDesc;
import com.sun.javadoc.ClassDoc;
import com.sun.javadoc.FieldDoc;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class TagTypeBuilder {

    private static final Logger LOG = LoggerFactory.getLogger(TagTypeBuilder.class);

    private ClassDoc classDoc;
    private File srcDir;
    private ObjectFactory objectFactory;

    public TagTypeBuilder(ClassDoc classDoc, File srcDir) {
        this.classDoc = classDoc;
        this.srcDir = srcDir;
        objectFactory = new ObjectFactory();
        LOG.debug("created a new TagTypeBuilder for class: {}", classDoc.qualifiedTypeName());
    }

    public TagType build() throws TldGeneratorException {
        TagType tagType = objectFactory.createTagType();
        AnnotationDesc jspTagAnnotation = JavaDocHelper.getAnnotation(classDoc, JspTag.class.getSimpleName());
        setTagTypeAttributes(tagType, jspTagAnnotation);
        setTldAttributeTypes(tagType);
        setVariableTypes(tagType);
        return tagType;
    }

    private void setTagTypeAttributes(TagType tagType, AnnotationDesc jspTagAnnotation) throws TldGeneratorException {
        // set the name
        String name = (String) JavaDocHelper.getAnnotationElementValue(jspTagAnnotation, "name");
        if (StringUtils.isNullOrEmpty(name)) {
            name = TagLibraryUtils.generateDefaultTagName(classDoc.typeName());
        }
        TldCanonicalNameType xmlName = objectFactory.createTldCanonicalNameType();
        xmlName.setValue(name);
        tagType.setName(xmlName);
        // set the description
        String description = classDoc.commentText();
        if (StringUtils.isNotNullOrEmpty(description)) {
            List<DescriptionType> descriptions = tagType.getDescriptions();
            DescriptionType xmlDescription = objectFactory.createDescriptionType();
            xmlDescription.setValue(description);
            descriptions.add(xmlDescription);
        }
        // set the tag class
        FullyQualifiedClassType tagClassType = objectFactory.createFullyQualifiedClassType();
        tagClassType.setValue(classDoc.qualifiedTypeName());
        tagType.setTagClass(tagClassType);
        // set the body content
        FieldDoc fieldDocBodyContent = (FieldDoc) JavaDocHelper.getAnnotationElementValue(jspTagAnnotation, "bodyContent");
        if (fieldDocBodyContent != null) {
            String bodyContentName = fieldDocBodyContent.name();
            if (StringUtils.isNotNullOrEmpty(bodyContentName)) {
                BodyContentType bodyContentType = objectFactory.createBodyContentType();
                bodyContentType.setValue(bodyContentName.toLowerCase());
                tagType.setBodyContent(bodyContentType);
            }
        }
        // set the dynamic attributes, the default is false
        Boolean dynamicAttributes = (Boolean) JavaDocHelper.getAnnotationElementValue(jspTagAnnotation, "dynamicAttributes");
        if ((dynamicAttributes != null) && dynamicAttributes) {
            GenericBooleanType xmlDynamicAttributes = objectFactory.createGenericBooleanType();
            xmlDynamicAttributes.setValue(dynamicAttributes.toString());
            tagType.setDynamicAttributes(xmlDynamicAttributes);
        }
        // set the example
        String example = (String) JavaDocHelper.getAnnotationElementValue(jspTagAnnotation, "example");
        if (StringUtils.isNotNullOrEmpty(example)) {
            XsdStringType xmlExample = objectFactory.createXsdStringType();
            xmlExample.setValue(example);
            tagType.setExample(xmlExample);
        }
        // set the TEI class
        String teiClassName = (String) JavaDocHelper.getAnnotationElementValue(jspTagAnnotation, "teiClass");
        if (StringUtils.isNullOrEmpty(teiClassName)) {
            try {
                teiClassName = TagLibraryUtils.findTeiClass(classDoc.qualifiedTypeName(), srcDir);
            } catch (IOException e) {
                throw new TldGeneratorException(e);
            }
        }
        if (StringUtils.isNotNullOrEmpty(teiClassName)) {
            FullyQualifiedClassType xmlTeiClass = objectFactory.createFullyQualifiedClassType();
            xmlTeiClass.setValue(teiClassName);
            tagType.setTeiClass(xmlTeiClass);
        }
    }

    private void setTldAttributeTypes(TagType tagType) throws TldGeneratorException {
        TldAttributeTypeListBuilder attributeBuilder = new TldAttributeTypeListBuilder(classDoc);
        List<TldAttributeType> attributesFromAnnotations = attributeBuilder.build();
        List<TldAttributeType> attributesFromTagType = tagType.getAttributes();
        for (TldAttributeType attribute : attributesFromAnnotations) {
            attributesFromTagType.add(attribute);
        }
        Collections.sort(attributesFromTagType, new Comparator<TldAttributeType>() {

            @Override
            public int compare(TldAttributeType t1, TldAttributeType t2) {
                JavaIdentifierType name1 = t1.getName();
                JavaIdentifierType name2 = t2.getName();
                return CompareToHelper.compareTo(name1, name2, "value");
            }

        });
    }

    private void setVariableTypes(TagType tagType) throws TldGeneratorException {
        VariableTypeListBuilder variableBuilder = new VariableTypeListBuilder(classDoc);
        List<VariableType> variablesFromAnnotations = variableBuilder.build();
        List<VariableType> variablesFromTagType = tagType.getVariables();
        for (VariableType variable : variablesFromAnnotations) {
            variablesFromTagType.add(variable);
        }
        Collections.sort(variablesFromTagType, new Comparator<VariableType>() {

            @Override
            public int compare(VariableType vt1, VariableType vt2) {
                JavaIdentifierType name1 = vt1.getNameFromAttribute();
                JavaIdentifierType name2 = vt2.getNameFromAttribute();
                int result = CompareToHelper.compareTo(name1, name2, "value");
                if (result == 0) {
                    name1 = vt1.getNameGiven();
                    name2 = vt2.getNameGiven();
                    result = CompareToHelper.compareTo(name1, name2, "value");
                }
                return result;
            }

        });
    }

}
